"""beaker-jobwatch - a tool to monitor beaker jobs and reschedule them."""

import datetime
import functools
import optparse
import os
import re
import select
import signal
import subprocess
import sys
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request
import xmlrpc.client

import lxml.etree

__version__ = "8.0"


OPTIONS: optparse.Values


# xml-rpc client for getting Beaker recipe status without having to fetch whole results
XMLRPC_CLIENT = xmlrpc.client.ServerProxy("https://beaker.engineering.redhat.com")


def readOptions(
    param_options,
    ignore_list,
):
    """read jobs to run."""

    herd = []
    rs_list = []
    duration = {}

    # populate herd with all recipesets from all given jobs (from both test-run and cmdline)
    for job in set(re.split(r"\W+", ",".join(param_options.jobs))):
        # if input was in TJ#XYZ format, if was split as [TJ,XYZ] - skip the TJ part
        # ditto for J:XYZ and empty jobs
        if job == "TJ" or job == "J" or job == "":
            continue
        print(f"processing job TJ#{job}")
        recipe_sets = RecipeSetCache.add_job(job)
        for recipe_set_id in recipe_sets:
            # check whether was the recipeSet already processed
            if recipe_set_id in rs_list:
                if OPTIONS.debug:
                    print(f"already processed recipeSet {recipe_set_id}")
            else:
                if OPTIONS.debug:
                    print(f"watching recipeSet {recipe_set_id}")
                rs = RecipeSet(
                    recipe_set_id,
                    job,
                    param_options.max_reschedules,
                    param_options.max_reschedules,
                    param_options.skip_destructive,
                    param_options.skip_machine,
                    ignore_list,
                )
                if rs.is_nacked() and not param_options.load_nacked:
                    if OPTIONS.debug:
                        print("recipeSet already NACKed, skipping")
                else:
                    # first recipe for this arch, note the time
                    if rs.description not in duration:
                        duration[rs.description] = {"start": datetime.datetime.now(), "end": None}
                    herd.append(rs)
                    # add RS ID to list of already parsed RS
                    rs_list.append(recipe_set_id)
                    if OPTIONS.debug:
                        print(f"newly processed recipeSet {recipe_set_id}")

    if len(herd) == 0:
        print("no recipeSets found (all nacked?)")
        sys.exit(1)

    print(f"\n{len(rs_list)} new recipeSet(s) added")

    return len(rs_list)


class Alarm(Exception):
    """support class/function for timeouting bkr calls"""

    @staticmethod
    def alarm_handler(signum, frame):
        """the actual handler fnc"""
        raise Alarm


class Bkr:
    """wrapper for bkr cmdline client calls"""

    username = None
    password = None

    class MaxRetriesReached(Exception):
        pass

    @classmethod
    def run(cls, args):
        cmd = ["bkr"]
        cmd.append(args[0])
        if cls.username and cls.password:
            cmd.extend(["--username=%s" % cls.username, "--password=%s" % cls.password])
        cmd.extend(args[1:])
        process = subprocess.Popen(args=cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return process

    @classmethod
    def timeout(cls, process, timeout=600):
        """Timeout for bkr calls."""
        out = None
        err = None
        signal.signal(signal.SIGALRM, Alarm.alarm_handler)
        signal.alarm(timeout)
        try:
            (out, err) = process.communicate()
            signal.alarm(0)  # reset alarm
        except Alarm:
            if OPTIONS.debug:
                print(f"timeout of {timeout} reached, terminating process {process.pid}")
            try:
                process.terminate()
                time.sleep(5)
                if process.poll() is None:
                    process.kill()
            except Exception:
                if OPTIONS.debug:
                    print("terminating process failed")
            time.sleep(10)
        return (out, err)

    @classmethod
    def run_repeatedly(cls, command, max_retries=5, pause=60):
        """runs command until it exits with 0 or we run out of retries"""
        retries = 0
        while retries < max_retries:
            p = cls.run(command)
            (output, error) = Bkr.timeout(p)
            if p.wait() == 0:
                return (output, error)
            print(f"bkr command {command} failed (try #{retries}):")
            print(error)
            time.sleep(pause)
            pause += 60
            retries += 1
        raise Bkr.MaxRetriesReached()


class LogCache:
    """stores log content from beaker recipes"""

    # usage logs['log_url']=log_content
    logs = {}
    # usage: log_list['recipe_id']=list_of_log_urls
    log_list = {}

    @classmethod
    def get_log(cls, log_url):
        """returns content of one log identified
        by a log URL (from cache if available)"""
        # check cache, update if empty
        if log_url not in cls.logs:
            if OPTIONS.debug:
                print(f"downloading log {log_url}")
            response = urllib.request.urlopen(log_url)
            cls.logs[log_url] = response.read().decode("utf-8")
        return cls.logs[log_url]

    # returns content of all logs for a particular recipe that matches log_name
    # (from cache if available)
    @classmethod
    def get_recipe_logs(cls, recipe_id, log_name="install.log"):
        # check cache, update if empty
        if recipe_id not in cls.log_list:
            if OPTIONS.debug:
                print(f"getting log list for recipe R:{recipe_id}")
            p = Bkr.run(["job-logs", f"R:{recipe_id}"])
            (output, err) = Bkr.timeout(p)
            output = output.decode()
            err = err.decode()
            if p.wait() != 0:
                print(f"retrieving recipe log list failed for R:{recipe_id}")
                return []
            cls.log_list[recipe_id] = list(str(output).split("\n"))
        log_list = [x for x in cls.log_list[recipe_id] if log_name in x]
        logs = []
        for log_url in log_list:
            try:
                logs.append(cls.get_log(log_url))
            except Exception as e:
                print(f"failed to retrieve log {log_url}:{e}")
        return logs


class RecipeSetCache:
    """Stores recipeset results in a parsed xml as returned by bkr job-results.

    Automatically updates cache for running RSs if older than defined interval.
    """

    cache = {}
    update_interval = 60

    @classmethod
    def add(cls, recipe_set_id):
        xml = cls._getResults(recipe_set_id)
        # keep whiteboard if we already have it
        try:
            job_whiteboard = cls.cache[recipe_set_id]["job_whiteboard"]
        except KeyError:
            job_whiteboard = "Unknown"
        try:
            first_seen = cls.cache[recipe_set_id]["first_seen"]
        except KeyError:
            first_seen = datetime.datetime.now()
        cls.cache[recipe_set_id] = {
            "xml": xml,
            "time": time.time(),
            "job_whiteboard": job_whiteboard,
            "first_seen": first_seen,
        }

    @classmethod
    def add_job(cls, job_id):
        """
        Fetches the entire job instead of individual RSs
        because there's no API to get RS from a job :(

        returns list of recipeset Ids found
        """
        xml = cls._getResults(job_id, recipe_set=False)
        job_whiteboard = xml.find("whiteboard").text
        rs_ids = []
        for rs in xml.findall(".//recipeSet"):
            rs_id = rs.get("id")
            rs_ids.append(rs_id)
            try:
                first_seen = cls.cache[rs_id]["first_seen"]
            except KeyError:
                first_seen = datetime.datetime.now()
            cls.cache[rs_id] = {
                "xml": rs,
                "time": time.time(),
                "job_whiteboard": job_whiteboard,
                "first_seen": first_seen,
            }
        return rs_ids

    @classmethod
    def get(cls, recipe_set_id, force=False, cache_only=False):
        if recipe_set_id not in cls.cache:
            return None
        if not force:
            # results are still fresh
            if (time.time() - cls.cache[recipe_set_id]["time"]) < cls.update_interval:
                return cls.cache[recipe_set_id]["xml"]
            if cache_only:
                return cls.cache[recipe_set_id]["xml"]
        # RS is finished, no update necessary
        running = False
        for recipe in cls.cache[recipe_set_id]["xml"].findall(".//recipe"):
            if recipe.get("status") not in ("Completed", "Cancelled", "Aborted"):
                running = True
        if running:
            cls.add(recipe_set_id)
        return cls.cache[recipe_set_id]["xml"]

    # an alias for the get method
    update = get

    @classmethod
    def _getResults(cls, object_id, recipe_set=True):
        prefix = "RS"
        if not recipe_set:
            prefix = "J"
        retries = 0
        while True:
            if OPTIONS.debug:
                print(f"fetching xml results of {prefix}:{object_id}")
                # traceback.print_stack()
            p = Bkr.run(["job-results", "--no-logs", "%s:%s" % (prefix, object_id)])
            (output, err) = Bkr.timeout(p)
            if p.wait() == 0:
                break
            print(f"retrieving {prefix} results failed for {prefix}:{object_id} (try #{retries})")
            retries += 1
            time.sleep(120)
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        return lxml.etree.XML(output, parser)

    @classmethod
    def first_seen(cls, recipe_set_id):
        """returns timestamp of when we added
        this recipe-set into the cache"""
        return cls.cache[recipe_set_id]["first_seen"]

    @classmethod
    def whiteboard(cls, recipe_set_id):
        """
        return job whiteboard for this RS
        """
        return cls.cache[recipe_set_id]["job_whiteboard"]

    @classmethod
    def dumpToDisk(cls, recipe_set_id):
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(cls.cache[recipe_set_id]["xml"]))
        file.close()
        print(f"RS {recipe_set_id} dumped to {filename}")


class RecipeSet:
    def __init__(
        self, id, job, reschedules, max_reschedules, skip_destructive, skip_machine, ignore_list=[]
    ):
        self.id = id
        self.job = job
        self.reschedules_left = reschedules
        self.max_reschedules = max_reschedules
        self.status_description = None
        self.active = True
        self.skip_destructive = skip_destructive
        self.skip_machine = skip_machine
        self.description = ""
        self.ignore_list = ignore_list
        xml = self.get_xml()
        for recipe in xml.findall(".//recipe"):
            self.description = "%s,%s" % (self.description, recipe.get("arch"))
        self.description = self.description.lstrip(",")
        if OPTIONS.debug:
            print(f"new recipeSet created (id={id}, job={job}, reschedules_left={reschedules}, "
                  f"ignore_list={ignore_list})")
        self.last_ping = {}
        self.last_reboot = {}
        self.reboot_time = 60 * OPTIONS.reboot_time
        self.cur_reboots_without_ping = {}

    def __str__(self):
        return str(self.id)

    def get_xml(self, cache_only: bool = True):
        """
        Return the parsed xml result of the entire RS
        """
        if OPTIONS.aggressive:
            cache_only = False
        return RecipeSetCache.get(self.id, cache_only=cache_only)

    def is_ok(self):
        """
        if no error was found (so far) in this recipeSet.
        Equals to is_finished unless in aggressive mode in which case
        the full results XML is fetched and analyzed.
        """
        # queued for longer than allowed
        if OPTIONS.queue_timeout:
            task_info = self.get_task_info()
            if task_info and task_info["state"] == "Queued":
                tracked_for = (
                    datetime.datetime.now() - RecipeSetCache.first_seen(self.id)
                ).total_seconds()
                if OPTIONS.debug:
                    print(f"RS:{self.id} first seen {RecipeSetCache.first_seen(self.id)}, "
                          f"remaining life: {OPTIONS.queue_timeout - tracked_for}")
                if tracked_for > OPTIONS.queue_timeout:
                    self.status_description = "queue timeout"
                    return False

        # reboots
        if OPTIONS.reboot:
            for box in self.get_machines():
                rc = os.system(f"ping -c 1 {box}")
                if os.WEXITSTATUS(rc) == 0:
                    if OPTIONS.debug:
                        print(f"Successful ping to {box} ( J:{self.job} / RS:{self.id} )")
                    self.last_ping[box] = time.time()
                    self.cur_reboots_without_ping[box] = 0
                else:
                    if (box in list(self.last_reboot.keys())) and (
                        (time.time() - self.last_reboot[box]) < self.reboot_time
                    ):
                        if OPTIONS.debug:
                            print(f"Waiting for {box} to reboot ( J:{self.job} / RS:{self.id} )")
                        continue
                    if OPTIONS.debug:
                        print(f"Unsuccessful ping to {box} ( J:{self.job} / RS:{self.id} )")
                # consider rebooting
                if box not in self.last_ping:
                    # here we assume that the box didn't install yet
                    if OPTIONS.debug:
                        print(f"Not a single ping to {box} yet ( J:{self.job} / RS:{self.id} ). "
                              f"Skipping it for now.")
                    continue
                if time.time() - self.last_ping[box] > self.reboot_time:
                    if (
                        self.cur_reboots_without_ping[box]
                        >= OPTIONS.max_reboots_without_ping
                    ):
                        print(f"System {box} fails to ping after "
                              f"{self.cur_reboots_without_ping[box]} reboots, rescheduling "
                              f"J:{self.job} / RS:{self.id}.")
                        # Cause a reschedule() and add the new job to herd (!)
                        self.status_description = f"failed to ping {box} after reboot"
                        return False
                    print(f"Rebooting {box} ( J:{self.job} / RS:{self.id} ) now ...")
                    os.system(f"bkr system-power --action reboot {box}")
                    self.last_reboot[box] = time.time()
                    self.cur_reboots_without_ping[box] += 1
                    continue

        # to decrease load on Beaker, don't fetch full xml
        if not OPTIONS.aggressive:
            if not self.is_finished():
                return True

        rs_xml = self.get_xml(cache_only=False)

        # destructive task aborting all tasks following it
        if "destructive-task" not in self.ignore_list:
            destructive_tasks = self.get_destructive_tasks()
            if destructive_tasks != []:
                self.status_description = "destructive task(s) %s" % (
                    functools.reduce(
                        lambda x, y: "%s,%s" % (x, y),
                        [task.get("name") for task in destructive_tasks],
                    )
                )
                return False

        # aborted/paniced recipes
        if "abort" not in self.ignore_list:
            for recipe in rs_xml.findall(".//recipe"):
                if recipe.get("status") == "Aborted":
                    # only check for panics in aborted recipes
                    # panics in completed ones are task-errors but not
                    # job breakers = ok, no need to reschedule
                    if recipe.get("result") == "Panic":
                        self.status_description = "kernel panic"
                    else:
                        self.status_description = "recipe aborted"
                        # check if no HW available for the recipe
                        results = recipe.xpath("//result")
                        if results:
                            if "does not match any systems" in results[0].text:
                                self.status_description = "no systems match required HW"
                                # irrecoverable error, don't reschedule
                                self.reschedules_left = 0
                        # check install logs for common errors
                        # check for broken scriptlets
                        logs = LogCache.get_recipe_logs(
                            recipe_id=recipe.get("id"), log_name="install.log"
                        )
                        if len(logs) > 0:
                            for line in logs[0].split("\n"):
                                match = re.match(r"^error:.*\((.*)\) scriptlet failed.*", line)
                                if match:
                                    package_name = match.group(1)
                                    self.status_description = (
                                        "bad scriptlet in %s package" % package_name
                                    )
                                    # irrecoverable error, don't reschedule
                                    self.reschedules_left = 0
                        else:
                            if OPTIONS.debug:
                                print(f"no install.log found for recipe R:{recipe.get('id')}")
                        # check for packaging issue (conflicts, missing deps,....)
                        packaging_issues = [
                            {
                                "file": "anaconda.log",
                                "pattern": r".*file conflicts.*installs of ([^ ]*) and ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*packaging: ([^ ]*) conflicts with ([^ ]*).*",
                                "message": "File conflict between {0} and {1}",
                            },
                            {
                                "file": "packaging.log",
                                "pattern": r".*WARN packaging: ([^ ]*) requires ([^ ]*)",
                                "message": "Package {0} has missing dependency: {1}",
                            },
                        ]
                        for issue in packaging_issues:
                            logs = LogCache.get_recipe_logs(
                                recipe_id=recipe.get("id"), log_name=issue["file"]
                            )
                            if len(logs) > 0:
                                for line in logs[0].split("\n"):
                                    if (match := re.match(issue["pattern"], line)):
                                        package_name1 = match.group(1)
                                        package_name2 = match.group(2)
                                        self.status_description = issue["message"].format(
                                            package_name1, package_name2
                                        )
                                        # packaging issues are irrecoverable, don't reschedule
                                        self.reschedules_left = 0
                            elif OPTIONS.debug:
                                print(f"no {issue['file']} found for recipe R:{recipe.get('id')}")
                    # only report as aborted if last task is aborted
                    # otherwise single task in the middle can make recipe aborted
                    # eventhough it's mostly usable [BZ#1327991]
                    last_task = rs_xml.findall(".//task")[-1]
                    if last_task.get("status") == "Aborted":
                        return False

        # failed /distribution/install
        if "install" not in self.ignore_list:
            for task in rs_xml.findall(".//task"):
                if task.get("name") == "/distribution/install":
                    if (task.get("result") not in ("Pass", "Warn", "New")) or (
                        task.get("status") in ("Aborted")
                    ):
                        self.status_description = "installation failed"
                        return False

        # critical task failure
        for task in rs_xml.findall(".//task"):
            if task.get("name") in OPTIONS.critical_tasks:
                if (
                    task.get("result") not in {"Pass", "New"}
                    and task.get("status") not in {"Cancelled"}
                ) or task.get("status") in {"Aborted"}:
                    self.status_description = "critical task %s failed" % task.get("name")
                    # irrecoverable error, don't reschedule
                    if task.get("name") in {
                        "/distribution/errata/setup",
                        "/distribution/errata/prepare",
                    }:
                        self.reschedules_left = 0
                        self.status_description = (
                            "critical task %s failed (irrecoverable)" % task.get("name")
                        )
                    return False

        # external watchdog
        if "watchdog" not in self.ignore_list:
            for task in rs_xml.findall(".//result"):
                if task.text == "External Watchdog Expired":
                    if task.getparent().getparent().get("name") == "/distribution/reservesys":
                        if OPTIONS.debug:
                            print(f"ignoring external watchdog in reservesys (RS:{self.id})")
                        continue
                    self.status_description = "killed by external watchdog"
                    return False

        return True

    def get_machines(self):
        """
        get list of machines on which this RS ran
        """
        rs_xml = self.get_xml()
        machines = []

        for machine in rs_xml.findall(".//recipe"):
            if machine.get("system") is not None:
                machines.append(machine.get("system"))
        return machines

    def get_destructive_tasks(self):
        """
        find if there is an aborted task and is followed only by other aborts
        """
        rs_xml = self.get_xml()

        first_aborted_task = None
        last_notaborted_task = None
        abort_count = 0
        results = []
        for task in rs_xml.findall(".//task"):
            if task.get("name") not in ("/distribution/install", "/distribution/reservesys"):
                if first_aborted_task is None:
                    if task.get("status") == "Aborted":
                        first_aborted_task = task
                    else:
                        last_notaborted_task = task
                else:
                    if task.get("status") != "Aborted":
                        first_aborted_task = None
                        last_notaborted_task = task
                        abort_count = 0
                    else:
                        abort_count += 1
        # last_notaborted_task == None => everything is aborted including install => no destrtask
        # first_aborted_task == None => nothing is aborted => no destrtask
        if last_notaborted_task is not None and first_aborted_task is not None:
            # if the task has at least one OK result then it's the culprit. Otherwise
            # it might be its predecessor - the last_notaborted_task
            for result in first_aborted_task.findall(".//result"):
                if not (
                    "LOCALWATCHDOG" in result.get("path")
                    or "Watchdog" in result.text
                    or "Task aborted by task" in result.text
                ):
                    results = [first_aborted_task]
            results = [first_aborted_task, last_notaborted_task]

        # if a destructive task is a setup task then we need to keep it [BZ#1331308]
        new_results = []
        for task in results:
            is_setup = False
            params = task.findall(".//param")
            for param in params:
                if param.get("name") == "WOW_TASK_TYPE" and param.get("value") == "setup":
                    if OPTIONS.debug:
                        print(
                            "not marking {} as destructive because it's a setup task".format(
                                task.get("name")
                            )
                        )
                    is_setup = True
            if not is_setup:
                new_results.append(task)
        results = new_results

        return results

    def get_task_info(self):
        """get the task_info object via XMLRPC"""
        tries = 3
        task_info = None
        signal.signal(signal.SIGALRM, Alarm.alarm_handler)
        while tries > 0:
            try:
                # Wait up to 60 seconds for xml client to return
                signal.alarm(60)
                task_info = XMLRPC_CLIENT.taskactions.task_info(f"RS:{self.id}")
                # Cancel alarm
                signal.alarm(0)
                break
            except Alarm:
                if OPTIONS.debug:
                    print("timeout of get_task_info")
            except Exception:
                # Cancel alarm
                signal.alarm(0)
                time.sleep(10)
                tries -= 1
        # restore the default handler
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
        return task_info

    # if the recipeSet is no longer running in beaker
    def is_finished(self) -> bool:
        # unless in aggressive mode, don't fetch the whole xml
        if not OPTIONS.aggressive:
            task_info = self.get_task_info()
            return task_info and task_info["is_finished"]

        # aggressive mode: heavy-weight check using the full xml result
        finished = True
        rs_xml = self.get_xml()
        # check recipeset status
        for recipe in rs_xml.findall(".//recipe"):
            if recipe.get("status") in {
                "Running",
                "Queued",
                "New",
                "Scheduled",
                "Processed",
                "Waiting",
                "Installing",
            }:
                finished = False
        # check all end-tasks in recipeset
        for task in rs_xml.findall(".//task"):
            if task.get("name") in OPTIONS.end_tasks:
                if task.get("result") not in ("New"):
                    finished = True
        if finished and self.is_ok():
            self.status_description = "ok"
        return finished

    def is_nacked(self):
        rs_xml = self.get_xml()
        if rs_xml.get("response") == "nak":
            return True
        return False

    def nack(self):
        if OPTIONS.debug:
            print("nacking RS:%s" % self.id)
        p = Bkr.run(["job-modify", "-r", "nak", "RS:%s" % self.id])
        if p.wait() != 0:
            print("nacking RS:%s failed" % self.id)
            return False
        return True

    def cancel(self):
        if OPTIONS.debug:
            print("canceling RS:%s" % self.id)
        p = Bkr.run(
            [
                "job-cancel",
                "RS:%s" % self.id,
                "--msg",
                "Canceled by beaker-jobwatch: %s" % self.status_description,
            ]
        )
        if p.wait() != 0:
            print("canceling RS:%s failed" % self.id)
            return False
        return True

    def reschedule(self):
        if self.reschedules_left == 0:
            print("recipeSet reached maximum number of reschedules")
            return -1, None, None
        # download vanilla xml job definition
        try:
            (output, error) = Bkr.run_repeatedly(
                ["job-clone", "--xml", "--dryrun", "RS:%s" % self.id]
            )
        except Bkr.MaxRetriesReached:
            print("getting job definition of RS:%s for cloning failed" % self.id)
            return -1, None, None
        parser = lxml.etree.XMLParser(remove_blank_text=True)
        job_xml = lxml.etree.XML(output, parser)

        # update whiteboard
        whiteboard = job_xml.find("whiteboard")
        reason = "clone of TJ#%s/RS:%s" % (self.job, self.id)
        whiteboard.text += "({} #{})".format(
            reason, 1 + self.max_reschedules - self.reschedules_left
        )

        # set group for the job
        if OPTIONS.job_group:
            job_xml.set("group", OPTIONS.job_group)
            if OPTIONS.debug:
                print("setting job group: {}".format(OPTIONS.job_group))

        # set owner for the job
        if OPTIONS.job_owner:
            job_xml.set("user", OPTIONS.job_owner)
            if OPTIONS.debug:
                print("setting job owner: {}".format(OPTIONS.job_owner))

        # skip destructive tasks
        if self.skip_destructive:
            destructive_tasks = self.get_destructive_tasks()
            if destructive_tasks != []:
                for destructive_task in destructive_tasks:
                    for task in job_xml.findall(".//task"):
                        if task.get("name") == destructive_task.get("name"):
                            # don't remove the last task
                            if task.getnext() is not None:
                                if OPTIONS.debug:
                                    print(
                                        "removing destructive task {} (id: {})".format(
                                            destructive_task.get("name"), destructive_task.get("id")
                                        )
                                    )
                                task.getparent().remove(task)
        # skip broken machine
        if self.skip_machine:
            for elem_hostrequire in job_xml.findall(".//hostRequires"):
                for machine in self.get_machines():
                    elem_and = lxml.etree.SubElement(elem_hostrequire, "and")
                    elem_hostname = lxml.etree.SubElement(elem_and, "hostname")
                    elem_hostname.set("op", "!=")
                    elem_hostname.set("value", machine)
                    if OPTIONS.debug:
                        print(f"excluding machine: {machine}")

        # dump to a file for submitting
        (fd, filename) = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(lxml.etree.tostring(job_xml))
        file.close()

        # submit to beaker as new job
        try:
            (output, error) = Bkr.run_repeatedly(["job-submit", filename])
            output = output.decode()
            error = error.decode()
        except Bkr.MaxRetriesReached:
            print("failed to submit cloned RS:%s" % (self.id))
            return -1, None
        match = re.match(r"^Submitted: \['J:([0-9]*)'\]", output)
        if not match:
            print("can't parse job id from job-clone output: %s" % output, end=" ")
            print("job definition: %s, stderr: %s" % (filename, error))
            return -1, None, None
        else:
            os.remove(filename)
            job = match.group(1)
            if OPTIONS.debug:
                print("new job TJ#%s scheduled" % job)
            recipe_set = RecipeSetCache.add_job(job)[0]
            if not self.is_finished():
                self.cancel()
            self.nack()
            self.active = False
            return int(job), reason, recipe_set

    def taskStatuses(self):
        """how many tasks are in particular status"""
        statuses = {
            "New": 0,
            "Processed": 0,
            "Queued": 0,
            "Scheduled": 0,
            "Waiting": 0,
            "Running": 0,
            "Completed": 0,
            "Cancelled": 0,
            "Aborted": 0,
        }
        for task in self.get_xml().findall(".//task"):
            statuses[task.get("status")] += 1
        return statuses

    def recipe_ids(self):
        """
        return list of all recipes that are part of this RS
        """
        recipes = self.get_xml().findall(".//recipe")
        recipe_ids = []
        for recipe in recipes:
            recipe_ids.append(recipe.get("id"))
        return recipe_ids

    def id_string(self) -> str:
        """
        Return a string representing this RS's ID.
        For a RS with a single recipe, return the recipe ID instead
        because it's more easily converted to URL (via doer)
        """
        if len(self.recipe_ids()) == 1:
            return "R:{}".format(self.recipe_ids()[0])
        return "RS:{}".format(self.id)


def main(args: list[str] | None = None) -> None:
    print(f"beaker-jobwatch v{__version__}")

    usage = """beaker-jobwatch watches beaker jobs and reschedules any recipe-sets that fail because
    of test-environment problems.

    beaker-jobwatch --job 127346[,TJ#1267,3575,...]
    """
    parser = optparse.OptionParser(usage=usage)
    parser.set_defaults(max_reschedules=3)
    parser.set_defaults(sleep=600)
    parser.add_option(
        "-j", "--job", dest="jobs", action="append", help="specify job(s) to be investigated"
    )
    parser.add_option(
        "-m",
        "--max",
        dest="max_reschedules",
        metavar="MAX",
        type="int",
        help="Maximum number of times one recipe set will be rescheduled",
    )
    parser.add_option(
        "",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Do not reschedule, cancel or nack failed recipesets, do not update run notes",
    )
    parser.add_option("", "--debug", dest="debug", action="store_true", help="Very verbose logging")
    parser.add_option(
        "",
        "--no-wait",
        dest="no_wait",
        action="store_true",
        help="Don't wait until running jobs finish",
    )
    parser.add_option(
        "",
        "--sleep",
        dest="sleep",
        type="int",
        help="""How long to wait before polling beaker for updated results (default: 600 seconds),\n
    pressing ENTER will skip the pause and force an update""",
    )
    parser.add_option(
        "",
        "--skip-destructive-task",
        dest="skip_destructive",
        action="store_true",
        help="Don't reschedule task(s) suspected of breaking the recipe",
    )
    parser.add_option(
        "",
        "--skip-broken-machines",
        dest="skip_machine",
        action="store_true",
        help="Instruct Beaker to not reuse machine(s) that ran the failed recipe set.",
    )
    parser.add_option(
        "",
        "--load-nacked",
        dest="load_nacked",
        action="store_true",
        help="Don't ignore nacked recipesets in specified jobs",
    )
    parser.add_option(
        "",
        "--ignore-failure",
        dest="ignore_list",
        action="append",
        default=["install"],
        help="Comma separated list of checks that will be ignored, these will not cause jobs to "
        "be canceled and rescheduled. Can be any combination of install, abort, watchdog and "
        "destructive-task or none. Default: install.",
    )
    parser.add_option(
        "", "--reboot", dest="reboot", action="store_true", help="Reboot unresponsive machines"
    )
    parser.add_option(
        "",
        "--reboot-time",
        dest="reboot_time",
        type="int",
        default=60,
        help="Reboot time treshold (defaults to 60 minutes)",
    )
    parser.add_option(
        "",
        "--max-reboots-without-ping",
        dest="max_reboots_without_ping",
        type="int",
        default=3,
        help="Reboot count treshold for machines not giving a ping",
    )
    parser.add_option(
        "",
        "--critical-task",
        dest="critical_tasks",
        action="append",
        default=[""],
        help="Failure of this task will trigger recipe cancellation, nacking and reschedule",
    )
    parser.add_option(
        "",
        "--end-task",
        dest="end_tasks",
        action="append",
        default=[""],
        help="Consider recipeset finished if an end task has completed, don't wait for the rest of "
        "the recipeset. Requires aggressive mode.",
    )
    parser.add_option(
        "",
        "--username",
        dest="username",
        help="Username for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_option(
        "",
        "--password",
        dest="password",
        help="Password for communication with beaker (instead of kerberos ticket)",
    )
    parser.add_option(
        "",
        "--job-group",
        dest="job_group",
        help="Reschedule jobs as group jobs belonging to given group",
    )
    parser.add_option(
        "", "--job-owner", dest="job_owner", help="Reschedule jobs as delegated to the given owner"
    )
    parser.add_option(
        "",
        "--aggressive",
        dest="aggressive",
        action="store_true",
        default=False,
        help="Fetch full results repeatedly, not just after recipe finishes. Stresses Beaker a "
        "lot, use with caution!",
    )
    parser.add_option(
        "",
        "--queue-timeout",
        dest="queue_timeout",
        type="int",
        default=None,
        help="Abort recipe set when queued for longer than defined (in seconds). Default: not set",
    )

    global OPTIONS  # pylint: disable=global-statement
    (OPTIONS, _) = parser.parse_args(args)

    if not OPTIONS.jobs:
        parser.print_help()
        sys.exit(1)

    # to evaluate end tasks, we need to fetch tasks results repeatedly
    # so turn aggressive mode on
    if len(OPTIONS.end_tasks) > 1:
        OPTIONS.aggressive = True
    if OPTIONS.debug:
        print("options: ", OPTIONS)

    if OPTIONS.debug and OPTIONS.reboot:
        print("reboot time:", OPTIONS.reboot_time, "min")

    if "BEAKER_USERNAME" in list(os.environ.keys()):
        Bkr.username = os.environ["BEAKER_USERNAME"]
    if "BEAKER_PASSWORD" in list(os.environ.keys()):
        Bkr.password = os.environ["BEAKER_PASSWORD"]
    if OPTIONS.username:
        Bkr.username = OPTIONS.username
    if OPTIONS.password:
        Bkr.password = OPTIONS.password

    ignore_list = []
    if OPTIONS.ignore_list:
        for option in ",".join(OPTIONS.ignore_list).split(","):
            if option not in ("install", "abort", "watchdog", "destructive-task", "none"):
                print(f"Invalid ignore-failure value {option}. Valid values are: "
                      f"install,abort,watchdog,destructive-task")
                sys.exit(1)
        ignore_list = ",".join(OPTIONS.ignore_list).split(",")
        if "none" in ignore_list:
            ignore_list = []

    RecipeSetCache.update_interval = OPTIONS.sleep - 1
    # list of successfully finished recipesets - will constitute results
    herd_finished_ok = []
    # list of unsuccessfully finished recipes
    herd_broken = []

    # dots are printed as a sort of status if nothing changes
    # lets count how many dots have we printed and add line break after 30
    dot_counter = 0

    # print status even if nothing changed since the last iteration, set to
    # True to print status at start
    force_status = True

    # good_recipesets: how many OK recipesets we must have to finish successfully
    # herd: list of recipesets we monitor - not finished yet
    # duration: measure how long does each arch take to finish
    (good_recipesets, herd, duration) = readOptions(OPTIONS, ignore_list)

    # main cycle iterating over all recipesets in herd until all are done =
    # = moved to other herds
    while True:
        fingerprint = str([herd, herd_finished_ok, herd_broken])

        # finished successfully
        new_herd = []
        for rs in herd:
            if rs.is_finished() and rs.is_ok():
                if OPTIONS.debug:
                    RecipeSetCache.dumpToDisk(rs.id)
                herd_finished_ok.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # broken (running or finished, doesn't matter)
        new_herd = []
        for rs in herd:
            if not rs.is_ok():
                if rs.reschedules_left > 0:
                    if not OPTIONS.dry_run:
                        new_job, reason, recipe_set_id = rs.reschedule()
                        if new_job > 0:
                            if OPTIONS.debug:
                                print("watching new recipeSet %s" % recipe_set_id)
                            herd.append(
                                RecipeSet(
                                    recipe_set_id,
                                    new_job,
                                    rs.reschedules_left - 1,
                                    OPTIONS.max_reschedules,
                                    rs.skip_destructive,
                                    rs.skip_machine,
                                    ignore_list,
                                )
                            )
                # normally cancellation happens after successful reschedule,
                # here we do it for recipes that won't be rescheduled (e.g. queue timeouts)
                elif not rs.is_finished():
                    rs.cancel()
                    rs.nack()
                    rs.active = False
                if OPTIONS.debug:
                    RecipeSetCache.dumpToDisk(rs.id)
                herd_broken.append(rs)
            else:
                new_herd.append(rs)
        herd = new_herd

        # print status
        status = ""
        new_fingerprint = str([herd, herd_finished_ok, herd_broken])
        if force_status or fingerprint != new_fingerprint:
            force_status = False
            # header
            print("\n")
            status += "%s\n" % datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            whiteboards = []
            for rs in herd + herd_finished_ok + herd_broken:
                whiteboard = RecipeSetCache.whiteboard(rs.id)
                # strip whiteboard of the postfix we add during rescheduling ("clone of ...")
                whiteboard = re.sub(r" *\(clone of.*\) *", "", whiteboard)
                # remove hyperlinks for Consistent identifier e.g.
                # [TR#330068](https://tcms.engineering.redhat.com/run/330068) -> [TR#330068]
                whiteboard = re.sub(r"(\[[^]]*\])\([^)]*\)", r"\g<1>", whiteboard)
                if whiteboard not in whiteboards:
                    status += f"{whiteboard}\n"
                    whiteboards.append(whiteboard)
            # individual herds
            status += f"Broken: {len(herd_broken)}\n"
            for rs in herd_broken:
                # indicate if max_reschedules has been reached
                # hard to spot from log if default (=3) wasn't used
                if rs.reschedules_left == 0:
                    attempts_string = "{} (max)".format(
                        OPTIONS.max_reschedules - rs.reschedules_left
                    )
                else:
                    attempts_string = str(OPTIONS.max_reschedules - rs.reschedules_left)
                status += "\tTJ#%s %s (%s): broken #%s: %s\n" % (
                    rs.job,
                    rs.id_string(),
                    rs.description,
                    attempts_string,
                    rs.status_description,
                )

            status += "Running:   %d/%d\n" % (len(herd), good_recipesets)
            for rs in herd:
                stats = rs.taskStatuses()
                tasks_finished = stats["Aborted"] + stats["Cancelled"] + stats["Completed"]
                tasks_all = functools.reduce(lambda x, y: x + y, list(stats.values()))
                # unless in aggressive mode we don't have up2date task info so ignore it
                if OPTIONS.aggressive:
                    tasks_stat_string = "({}/{})".format(tasks_finished, tasks_all)
                else:
                    tasks_stat_string = "({})".format(tasks_all)
                status += "\tTJ#%s %s (%s): running %s\n" % (
                    rs.job,
                    rs.id_string(),
                    rs.description,
                    tasks_stat_string,
                )

            status += "Completed: %d/%d\n" % (good_recipesets - len(herd), good_recipesets)
            if len(herd) + len(herd_finished_ok) == 0:
                print("no jobs in herd + herd_finished_ok, this should never happen")
                print("(unless you inspect finished broken jobs only)")
                print(["TJ#%s %s" % (rs.job, rs.id_string()) for rs in (herd)])
                print(["TJ#%s %s" % (rs.job, rs.id_string()) for rs in (herd_finished_ok)])
                print(["TJ#%s %s" % (rs.job, rs.id_string()) for rs in (herd_broken)])
            else:
                if len(herd_broken) + len(herd_finished_ok) != 0:
                    status += "\t%s\n" % functools.reduce(
                        lambda x, y: "%s %s" % (x, y),
                        {"TJ#%s" % rs.job for rs in (herd_broken + herd_finished_ok)},
                    )
                status += (
                    "https://beaker.engineering.redhat.com/matrix/?toggle_nacks_on=on&job_ids=%s"
                    % functools.reduce(
                        lambda x, y: "%s+%s" % (x, y),
                        {rs.job for rs in (herd + herd_finished_ok + herd_broken)},
                    )
                )
            print(status)

            # check if some arch isn't completely finished, note the time
            for arch in list(duration.keys()):
                if arch not in [x.description for x in herd] and duration[arch]["end"] is None:
                    duration[arch]["end"] = datetime.datetime.now()
        else:
            sys.stdout.write(".")
            dot_counter += 1
            if dot_counter == 30:
                print("\n%s" % datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
                dot_counter = 0
        sys.stdout.flush()

        if not OPTIONS.no_wait and len(herd) > 0:
            # sleep between updates, hitting ENTER will skip
            # the sleep and force job-results update
            if sys.stdin.isatty():
                slept = 0
                while slept < OPTIONS.sleep:
                    i, o, e = select.select([sys.stdin], [], [], 0.0001)
                    for s in i:
                        if s == sys.stdin:
                            sys.stdin.readline()  # read line = enter pressed
                            print("skipping pause and forcing update of all running recipe-sets")
                            slept = OPTIONS.sleep
                            force_status = True
                            for rs_id in {rs.id for rs in herd}:
                                RecipeSetCache.update(rs_id, force=True)
                            # flush the rest of input so we start clear the next round
                            while len(select.select([sys.stdin.fileno()], [], [], 0.0)[0]) > 0:
                                os.read(sys.stdin.fileno(), 4096)
                                # just in case we've lost tty to not block infinitely
                                if not sys.stdin.isatty():
                                    break
                    time.sleep(1)
                    slept += 1
            else:
                # not running on terminal, just sleep the whole time
                time.sleep(OPTIONS.sleep)
        else:
            break
    # dump herds for parsing out the (un)successful recipe sets
    if OPTIONS.debug:
        print(f"herd_broken: {','.join([str(x) for x in herd_broken])}")
        print(f"herd_finished_ok: {','.join([str(x) for x in herd_finished_ok])}")
    # print duration
    if not OPTIONS.no_wait:
        delta = max([duration[x]["end"] for x in duration]) - min(
            [duration[x]["start"] for x in duration]
        )
        print(f"duration: {delta}")
        if OPTIONS.debug:
            for arch in duration:
                print(f"duration of {arch}: {duration[arch]['end'] - duration[arch]['start']}")
    if False in [rs.is_ok() for rs in herd if rs.active] or good_recipesets > len(herd_finished_ok):
        print("finished unsuccessfully")
        sys.exit(2)

    print("finished successfully")
